#include <math.h>
#include <stdio.h> 
#include <string.h>

#include "brus_slv.h"
#include "utils.h"


/* Fuer Lambda_Hierarchical(4,1): ord:2  */
double           hj1, j10, j01, j00,j11,j011,j101,j110,j111,j1111; 
unsigned long    until_stationary;
random_set       gauss1;


 
/* Brusselator: Taylor: Lambda_hierarchical(4,1) */
 
void Brus_nonlinear_ord_2 (double x_start, double *x_new, 
                           double y_start, double *y_new,
                           double dt, double alpha, double beta, double sigma) {
 
        double t1,t2,t3,t5,t6,t7,t10,t11,t12,t16,t21,t22,t24,t27,t30,t33,t34,t38,t42,
                t44,t45,t47,t48,t49,t64;

        t1 = sigma*x_start;
        t2 = x_start*y_start;
        t3 = -beta-1.0+2.0*t2;
        t5 = t1*y_start;
        t6 = x_start*x_start;
        t7 = sigma*t6;
        t10 = sigma*sigma;
        t11 = t6*x_start;
        t12 = t10*t11;
        t16 = t10*x_start*t3;
        t21 = t6*y_start;
        t22 = alpha-(beta+1.0)*x_start+t21;
        t24 = j011*t22*t10;
        t27 = j111*t10*sigma*x_start;
        t30 = hj1*sigma*x_start;
        t33 = beta*x_start-t21;
        t34 = t33*t6;
        t38 = sigma*t11;
        t42 = j01*t22*sigma;
        t44 = j11*t10*x_start;
        t45 = t10*t10;
        t47 = j1111*t45*x_start;
        t48 = x_start+j110*(-t1*(-sigma*t3-2.0*t5+3.0*t7)-2.0*t12)+j101*(t16-t12)+
                t24-t27+dt*t22-t30+j00*(t22*t3+t34)+j10*(-t1*t3+t38)-t42+t44+t47;
        t49 = beta-2.0*t2;
        t64 = y_start+j110*(-t1*(-sigma*t49+2.0*t5-3.0*t7)+2.0*t12)+j101*(-t16+t12)-
                t24+t27+dt*t33+t30+j00*(t22*t49-t34)+j10*(-t1*t49-t38)+t42-t44-t47;
        *x_new = t48;
        *y_new = t64;
}

void Brus_nonlinear_ord2_fwd_for_grid (double *xini, double *yini, 
				       double dt, 
				       double alpha, double beta, double sigma) {
  
  double t1,t2,t3,t5,t6,t7,t10,t11,t12,t16,t21,t22,t24,t27,t30,t33,t34,t38,t42,
         t44,t45,t47,t48,t49,t64;
  double x_old, y_old;
  
  x_old = *xini;
  y_old = *yini;
  
  t1 = sigma*x_old;
  t2 = x_old*y_old;
  t3 = -beta-1.0+2.0*t2;
  t5 = t1*y_old;
  t6 = x_old*x_old;
  t7 = sigma*t6;
  t10 = sigma*sigma;
  t11 = t6*x_old;
  t12 = t10*t11;
  t16 = t10*x_old*t3;
  t21 = t6*y_old;
  t22 = alpha-(beta+1.0)*x_old+t21;
  t24 = j011*t22*t10;
  t27 = j111*t10*sigma*x_old;
  t30 = hj1*sigma*x_old;
  t33 = beta*x_old-t21;
  t34 = t33*t6;
  t38 = sigma*t11;
  t42 = j01*t22*sigma;
  t44 = j11*t10*x_old;
  t45 = t10*t10;
  t47 = j1111*t45*x_old;
  t48 = x_old+j110*(-t1*(-sigma*t3-2.0*t5+3.0*t7)-2.0*t12)+j101*(t16-t12)+
        t24-t27+dt*t22-t30+j00*(t22*t3+t34)+j10*(-t1*t3+t38)-t42+t44+t47;
  t49 = beta-2.0*t2;
  t64 = y_old+j110*(-t1*(-sigma*t49+2.0*t5-3.0*t7)+2.0*t12)+j101*(-t16+t12)-
        t24+t27+dt*t33+t30+j00*(t22*t49-t34)+j10*(-t1*t49-t38)+t42-t44-t47;
  *xini = t48;
  *yini = t64;
}


/* Linearized Brusselator, linearized in (x0,y0): Taylor: Lambda_hierarchical(4,1) */
 
void Brus_linear_ord_2 (double x_start, double x0, double *x_new, 
                        double y_start, double y0, double *y_new, 
                        double dt, double beta, double sigma) {

        double t1,t2,t3,t5,t6,t10,t11,t12,t13,t17,t18,t20,t23,t26,t28,
                t30,t31,t34,t36,t40,t42,t43,t45,t46,t60;
                

        t1 = j110*sigma;
        t2 = x0*y0;
        t3 = -beta-1.0+2.0*t2;
        t5 = x0*x0;
        t6 = sigma*t5;
        t10 = sigma*sigma;
        t11 = t10*x_start;
        t12 = t11*t3;
        t13 = t11*t5;
        t17 = t5*y_start;
        t18 = t3*x_start+t17;
        t20 = j011*t18*t10;
        t23 = j111*t10*sigma*x_start;
        t26 = hj1*sigma*x_start;
        t28 = beta-2.0*t2;
        t30 = t28*x_start-t17;
        t31 = t30*t5;
        t34 = sigma*x_start;
        t36 = t34*t5;
        t40 = j01*t18*sigma;
        t42 = j11*t10*x_start;
        t43 = t10*t10;
        t45 = j1111*t43*x_start;
        t46 = x_start-t1*x_start*(-sigma*t3+t6)+j101*(t12-t13)+t20-t23+dt*t18-t26+j00*(
                t18*t3+t31)+j10*(-t34*t3+t36)-t40+t42+t45;
        t60 = y_start-t1*x_start*(-sigma*t28-t6)+j101*(-t12+t13)-t20+t23+dt*t30+t26+j00
                *(t18*t28-t31)+j10*(-t34*t28-t36)+t40-t42-t45;
        *x_new = t46;
        *y_new = t60;
}


/* Brusselator with negative drift: Taylor: Lambda_hierarchical(4,1) */

void brus_nonlinear_ord2_bwd (double x_start, double *x_new, 
			      double y_start, double *y_new,
			      double dt, double alpha, double beta, double sigma) {

  double t1,t2,t4,t6,t9,t10,t11,t13,t15,t16,t17,
         t20,t21,t24,t26,t27,t31,t32,t36,t39,t42,
         t43,t48,t49,t64;

  t1 = sigma*sigma;
  t2 = t1*t1;
  t4 = j1111*t2*x_start;
  t6 = hj1*sigma*x_start;
  t9 = x_start*x_start;
  t10 = t9*y_start;
  t11 = -alpha+(beta+1.0)*x_start-t10;
  t13 = j01*t11*sigma;
  t15 = j11*t1*x_start;
  t16 = x_start*y_start;
  t17 = beta+1.0-2.0*t16;
  t20 = -beta*x_start+t10;
  t21 = t20*t9;
  t24 = sigma*x_start;
  t26 = t9*x_start;
  t27 = sigma*t26;
  t31 = t1*x_start*t17;
  t32 = t1*t26;
  t36 = j011*t11*t1;
  t39 = j111*t1*sigma*x_start;
  t42 = t24*y_start;
  t43 = t9*sigma;
  t48 = x_start+t4-t6-t13+t15+j00*(t11*t17-t21)+j10*(-t24*t17-t27)
        +j101*(t31+t32)+t36-t39+dt*t11+j110*(-t24*(-t17*sigma+2.0*t42-3.0*t43)+2.0*t32);
  t49 = -beta+2.0*t16;
  t64 = y_start-t4+t6+t13-t15+j00*(t11*t49+t21)+j10*(-t24*t49+t27)
        +j101*(-t31-t32)-t36+t39+dt*t20+j110*(-t24*(-sigma*t49-2.0*t42+3.0*t43)-2.0*t32);
  *x_new = t48;
  *y_new = t64;
}


/* Linearized Brusselator with negative drift, linearized in (x0,y0): 
   Taylor: Lambda_hierarchical(4,1) */

void brus_lin_ord2_bwd (double x_start, double x0, double *x_new, 
                        double y_start, double y0, double *y_new, 
                        double dt, double beta, double sigma) {

  double t1,t2,t4,t6,t7,t8,t10,t11,t12,t14,t16,t18,t20,t21,
         t24,t26,t29,t30,t31,t35,t38,t40,t42,t46,t60;

  t1 = sigma*sigma;
  t2 = t1*t1;
  t4 = j1111*t2*x_start;
  t6 = hj1*sigma*x_start;
  t7 = x0*y0;
  t8 = beta+1.0-2.0*t7;
  t10 = x0*x0;
  t11 = t10*y_start;
  t12 = t8*x_start-t11;
  t14 = j01*t12*sigma;
  t16 = j11*t1*x_start;
  t18 = -beta+2.0*t7;
  t20 = t18*x_start+t11;
  t21 = t20*t10;
  t24 = sigma*x_start;
  t26 = t24*t10;
  t29 = t1*x_start;
  t30 = t29*t8;
  t31 = t29*t10;
  t35 = j011*t12*t1;
  t38 = j111*t1*sigma*x_start;
  t40 = j110*sigma;
  t42 = t10*sigma;
  t46 = x_start+t4-t6-t14+t16+j00*(t12*t8-t21)+j10*(-t24*t8-t26)+
        j101*(t30+t31)+t35-t38+dt*t12-t40*x_start*(-t8*sigma-t42);
  t60 = y_start-t4+t6+t14-t16+j00*(t12*t18+t21)+j10*(-t24*t18+t26)+
        j101*(-t30-t31)-t35+t38+dt*t20-t40*x_start*(-sigma*t18+t42);
  *x_new = t46;
  *y_new = t60;
}



/* Brusselator rueckwaerts in der Zeit: Taylor: Lambda_hierarchical(4,1) */
 
void Brus_nonlinear_ord2_bwd_for_grid (double *xini, double *yini, 
				       double dt, 
				       double alpha, double beta, double sigma) {
        
        double t1,t2,t4,t6,t9,t10,t11,t13,t15,t16,t17,t20,t21,t24,
                t26,t27,t31,t32,t36,t39,t42,t43,t48,t49,t64;
        double x_old, y_old;
  
        x_old = *xini;
        y_old = *yini;


        t1 = sigma*sigma;
        t2 = t1*t1;
        t4 = j1111*t2*x_old;
        t6 = hj1*sigma*x_old;
        t9 = x_old*x_old;
        t10 = t9*y_old;
        t11 = -alpha+(beta+1.0)*x_old-t10;
        t13 = j01*t11*sigma;
        t15 = j11*t1*x_old;
        t16 = x_old*y_old;
        t17 = beta+1.0-2.0*t16;
        t20 = -beta*x_old+t10;
        t21 = t20*t9;
        t24 = sigma*x_old;
        t26 = t9*x_old;
        t27 = sigma*t26;
        t31 = t1*x_old*t17;
        t32 = t1*t26;
        t36 = j011*t11*t1;
        t39 = j111*t1*sigma*x_old;
        t42 = t24*y_old;
        t43 = t9*sigma;
        t48 = x_old+t4-t6-t13+t15+j00*(t11*t17-t21)+
                j10*(-t24*t17-t27)+j101*(t31+t32)+t36-t39+dt*t11+
                j110*(-t24*(-t17*sigma+2.0*t42-3.0*t43)+2.0*t32);
        t49 = -beta+2.0*t16;
        t64 = y_old-t4+t6+t13-t15+j00*(t11*t49+t21)+
                j10*(-t24*t49+t27)+j101*(-t31-t32)-t36+t39+dt*t20+
                j110*(-t24*(-sigma*t49-2.0*t42+3.0*t43)-2.0*t32);
        *xini = t48;
        *yini = t64;
}


void compute_j_functionals(random_set *ptr_rs, double time_step) {

/* Fuer Lambda_Hierarchical(4,1): ord:2  begin*/
  hj1  = j_1(ptr_rs,fabs(time_step)); 
  j10  = j_10(ptr_rs,fabs(time_step)); 
  j11  = j_11(ptr_rs,fabs(time_step)); 
  j00  = j_00(fabs(time_step)); 
  j01  = j_01(ptr_rs,fabs(time_step)); 
  j011  = j_011(ptr_rs,fabs(time_step)); 
  j101 = j_101(ptr_rs,fabs(time_step)); 
  j110 = j_110(ptr_rs,fabs(time_step)); 
  j111 = j_111(ptr_rs,fabs(time_step)); 
  j1111 = j_1111(ptr_rs,fabs(time_step)); 
  /*  Fuer Lambda_Hierarchical(4,1) end*/
  /*Fuer Lambda_Hierarchical(3,1): ord:1.5  begin 
    hj1  = j_1(ptr_rs,fabs(time_step)); 
    j01  = j_01(ptr_rs,fabs(time_step)); 
    j10  = j_10(ptr_rs,fabs(time_step)); 
    j11  = j_11(ptr_rs,fabs(time_step)); 
    j111  = j_111(ptr_rs,fabs(time_step)); 
    Fuer Lambda_Hierarchical(3,1)  end 
    */
  /*Fuer Lambda_Hierarchical(2,1): ord:1  begin 
    hj1  = j_1(ptr_rs,fabs(time_step)); 
    j11  = j_11(ptr_rs,fabs(time_step)); 
    Fuer Lambda_Hierarchical(2,1)  end 
    */
}

/* compute stationary nonlinear solution */
void  compute_stationary_nl_solution ( double alpha,
				       double beta,
				       double sigma,
				       double dt,
				       double nlx1_s,
				       double nlx2_s,
				       double nly1_s,
				       double nly2_s,
				       FILE   *fp_log,
				       int    modulus,
				       double *nlx_stat,
				       double *nly_stat
				       ) {
  
  int       stationary_nl_solution_reached;
  int       loop;
  double    diff_x,diff_y,arith_mean_x,arith_mean_y;
  double    nlx1_old=0.0,nlx2_old=0.0;
  double    nlx1_new=0.0,nlx2_new=0.0;
  double    nly1_old=0.0,nly2_old=0.0; 
  double    nly1_new=0.0,nly2_new=0.0;
  double    time;

/*
  if (DEBUG) {
    fprintf(stderr, "Computing stationary solution for BETA= %g\n\n", beta);
    fprintf(stderr,"time \t nlx1 \t\t nlx2 \t\t nly1 \t\t nly2\n");
  }
*/
  nlx1_old = nlx1_s; 
  nly1_old = nly1_s;
  nlx2_old = nlx2_s;
  nly2_old = nly2_s;
  srand48(UNTIL_STATIONARY_SEED);
  stationary_nl_solution_reached = FALSE;
/*
  if (DEBUG) {
    fprintf(stderr,"%g  \t%.6f\t%.6f\t%.6f\t%.6f\n",0.0,nlx1_old,nlx2_old,nly1_old,nly2_old);
  }
*/
  time = dt;
  loop = 1;
  while ( (!stationary_nl_solution_reached) || 
	  (loop <= until_stationary) ) {
    fill_gauss(&gauss1);
    compute_j_functionals(&gauss1,dt);	
    Brus_nonlinear_ord_2(nlx1_old, &nlx1_new, nly1_old, &nly1_new, 
			 dt, alpha, beta, sigma); 
    Brus_nonlinear_ord_2(nlx2_old, &nlx2_new, nly2_old, &nly2_new, 
			 dt, alpha, beta, sigma); 
    if ( (fabs(nlx1_new) > MAX_SDE) || (fabs(nly1_new) > MAX_SDE) ) {
      fprintf(fp_log, "Computation of the stationary nonlinear solution terminated\n");
      fprintf(fp_log, "at time = %g for beta = %g and the initial value (x(0),y(0))=(%g,%g)\n", 
	      time,beta,nlx1_s,nly1_s);
      fprintf(fp_log, "as |x(%g)| or |y(%g)| > %g\n",time,time,MAX_SDE);
      exit(2);
    }
    if ( (fabs(nlx2_new) > MAX_SDE) || (fabs(nly2_new) > MAX_SDE) ) {
      fprintf(fp_log, "Computation of the stationary nonlinear solution terminated\n");
      fprintf(fp_log, "at time = %g for beta = %g and the initial value (x(0),y(0))=(%g,%g)\n", 
	   time,beta,nlx2_s,nly2_s);
      fprintf(fp_log, "as |x(%g)| or |y(%g)| > %g\n",time,time,MAX_SDE);
      exit(2);
    }  
    diff_x = fabs(nlx1_new - nlx2_new);
    diff_y = fabs(nly1_new - nly2_new);
    arith_mean_x = fabs(nlx1_new + nlx2_new)*0.5;
    arith_mean_y = fabs(nly1_new + nly2_new)*0.5;
    if( diff_x < (EPS_DIFF_X*arith_mean_x) && diff_y < (EPS_DIFF_Y*arith_mean_y) ){
      stationary_nl_solution_reached = TRUE;
    }
/*
    if (DEBUG) {
      if ( (loop % modulus) == 0 ) {
	fprintf(stderr,"%g  \t%.6f\t%.6f\t%.6f\t%.6f\n",time,nlx1_new,nlx2_new,nly1_new,nly2_new);
	fflush(stderr);
      }
    }
*/
    nlx1_old = nlx1_new;
    nly1_old = nly1_new;
    nlx2_old = nlx2_new;
    nly2_old = nly2_new;
    time += dt;
    loop++;
  }
  *nlx_stat = nlx1_new;
  *nly_stat = nly1_new;
/*
  if (DEBUG) {
    fprintf(stderr,"The stationary solution for BETA=%g at time %g is:\n"
	    "%g  \t%.6f\t%.6f\t%.6f\t%.6f\n\n",
	    beta,time-dt,time-dt,nlx1_new,nlx2_new,nly1_new,nly2_new);
  }
*/
}

