#include <stdlib.h>
#include <math.h>
#include <stdio.h> 
#include "j.h"

#if defined(__MSDOS__) || defined(__WIN32__)
#ifndef __rand48__
#define __rand48__
#define drand48()	(rand()/(1.0+RAND_MAX))
#define srand48(s)	srand(s)
#endif
#endif

#ifndef M_SQRT1_2 
#define M_SQRT1_2 (1/sqrt(2.0)) 
#endif

void gauss_init(long seed)
{
  /* srand48(234); */
  /*srand48(951); */
  /*srand48(4314158); */
  srand48(seed); 
}


void fill_gauss(random_set* rs1)
{
  static double zwopi= 2.0 * M_PI;
  double zuf1,zuf2;
  long i;

  zuf1=drand48();
  zuf2=drand48();
  rs1->xi = sqrt(-2.0 * log(1.0-zuf1)) * cos(zwopi * zuf2);
  for(i=1;i<11;i++) {
    zuf1=drand48();
    zuf2=drand48();
      rs1->eta[i]  = sqrt(-2.0 * log(1.0-zuf1)) * cos(zwopi * zuf2) ;
      rs1->zeta[i] = sqrt(-2.0 * log(1.0-zuf1)) * sin(zwopi * zuf2) ;
  }
}

double a0(random_set* rs1, double dt)
{
  long i;
  double sum=0.0;

  for(i=1;i<11;i++)  sum += (rs1->zeta[i] / i);

  sum = -sqrt(2*dt) * M_1_PI * sum;
  return(sum);
}

double b0(random_set* rs1, double dt)
{
  long i;
  double sum=0.0;

  for(i=1;i<11;i++)  sum += (rs1->eta[i] / (i*i) );

  sum = M_SQRT1_2 * sqrt(dt) * sum;
  return(sum);
}

double A_11(random_set* rs1, double dt)
{
  return(0.0);
}

double A_12(random_set* rs1, random_set* rs2, double dt)
{
  long i;
  double sum=0.0;

  for(i=1;i<11;i++)
     sum += ((rs1->zeta[i]*rs2->eta[i] + rs1->eta[i]*rs2->zeta[i]) /i );

  sum = sum / (2 * M_PI);
  return(sum);
}

double B11(random_set* rs1, double dt)
{
  long i;
  double sum=0.0;

  for(i=1;i<11;i++)
     sum += ((rs1->eta[i]*rs1->eta[i] + rs1->zeta[i]*rs1->zeta[i]) / (i*i) );

  sum = sum / (4 * M_PI * M_PI);
  return(sum);
}

double B12(random_set* rs1, random_set* rs2, double dt)
{
  long i;
  double sum=0.0;

  for(i=1;i<11;i++)
     sum += ((rs1->eta[i]*rs2->eta[i] + rs1->zeta[i]*rs2->zeta[i]) / (i*i) );

  sum = sum / (4 * M_PI * M_PI);
  return(sum);
}

double C11(random_set* rs1, double dt)
{
  long i,j;
  double sum=0.0;

  for(i=1;i<11;i++)
    for(j=1;j<11;j++)
      if (i!=j) 
        sum+=(i* (  rs1->zeta[i]*rs1->zeta[j]/j  \
                     - rs1->eta[i] *rs1->eta[j] *j/i  ) / (i*i - j*j)  );
  sum = -0.5 * sum / (M_PI * M_PI);
  return(sum);
}

double C12(random_set* rs1, random_set* rs2, double dt)
{
  long i,j;
  double sum=0.0;

  for(i=1;i<11;i++)
    for(j=1;j<11;j++)
      if (i!=j) 
        sum+=(i*i * (  rs1->zeta[i]*rs2->zeta[j]/j  \
                     - rs1->eta[i] *rs2->eta[j] *j/i  ) / (i*i - j*j)  );

/*

xxxxxxxxxxx

war 

  sum+=(i*i * (  rs1->zeta[i]*rs2->zeta[j]/j  \
                     - rs1->eta[i] *rs2->eta[j] *j/r  ) / (i*i - j*j)  );





*/
  sum = -0.5 * sum / (M_PI * M_PI);
  return(sum);
}

double j_1(random_set* rs1, double dt)
{
  return(sqrt(dt) * rs1->xi);
}

double j_0(double dt)
{
  return(dt);
}

double j_11(random_set* rs1, double dt)
{
  double j1;

  j1 = j_1(rs1,dt);
  return(.5 * j1 * j1);
}

double j_12(random_set* rs1, random_set* rs2, double dt)
{
  double j1,j2,a1,a2,A12,sum;

  j1 = j_1(rs1,dt);
  j2 = j_1(rs2,dt);
  a1 = a0(rs1,dt);
  a2 = a0(rs2,dt);
  A12= A_12(rs1,rs2,dt);
  sum = .5 * (j1*j2 + a1*j2 - a2*j1) + dt*A12;

  return(sum);
}

double j_10(random_set* rs1, double dt)
{
  double j1,a;

  j1 = j_1(rs1,dt);
  a = a0(rs1,dt);
  return( .5 * dt * (j1 + a) );
}

double j_01(random_set* rs1, double dt)
{
  double j1,a;

  j1 = j_1(rs1,dt);
  a = a0(rs1,dt);
  return( .5 * dt * (j1 - a) );
}

double j_111(random_set* rs1, double dt)
{                                        
  double j1;

  j1 = j_1(rs1,dt);
  return(j1 * j1 * j1 / 6);
}

double j_00(double dt)
{
  return(.5 * dt * dt);
}

double j_101(random_set* rs1, double dt)
{
  double j1,a,b,B;

  j1 = j_1(rs1,dt);
  a = a0(rs1,dt);
  b = b0(rs1,dt);
  B = B11(rs1,dt);
  
  return( dt * (j1*(j1/6 + b/M_PI) - a*a/4 - B) );
}

double j_011(random_set* rs1, double dt)
{                             
  double j1,a,b,B,C;

  j1 = j_1(rs1,dt);
  a = a0(rs1,dt) / 4;
  b = b0(rs1,dt) * .5 / M_PI;
  B = B11(rs1,dt);
  C = C11(rs1,dt);
  
  return( dt * ( j1 * ( j1/6 - a - b) + dt*(B+C) ) );
}

double j_110(random_set* rs1, double dt)
{
  double j1,j101,j011;

  j1 = j_1(rs1,dt);
  j101 = j_101(rs1,dt);
  j011 = j_011(rs1,dt);
  return( dt * j1*j1 / 2 - j101 - j011 );
}

double j_1111(random_set* rs1, double dt)
{                                          
  double j1;

  j1 = j_1(rs1,dt);
  return(j1 * j1 * j1 * j1 / 24);
}
