#include <stdlib.h>
#include <stdio.h> 
#include <string.h>
#include <math.h>
#include <values.h>

#include "brus_slv.h"
#include "j.h"
#include "utils.h"

/*****  global variables  *****/

long            gauss_seed;
double          alpha; 
double          beta,beta_start,beta_end;
double          sigma;
double          dt;                    
double          t_end; 
unsigned long   iterations;
double          t_until_stationary;
double          lx_s,ly_s;
double          nlx1_s,nly1_s,nlx2_s,nly2_s;
double          **xnon, **ynon;
unsigned long   no_paramvalues;
double          delta_param;
char            *user_filename;
int             user_filenamesize;

/*****  global variables  *****/


void initialize(char *argv[]) {

  sscanf(argv[1],  "%d",  &user_filenamesize);
  user_filename = salloc((user_filenamesize+1)*sizeof(*user_filename));
  sscanf(argv[2],  "%s",  user_filename);
  sscanf(argv[3],  "%ld", &gauss_seed);
  sscanf(argv[4],  "%lf", &t_end);
  sscanf(argv[5],  "%lf", &dt);
  sscanf(argv[6],  "%lf", &lx_s); 
  sscanf(argv[7],  "%lf", &ly_s); 
  sscanf(argv[8],  "%lf", &nlx1_s); 
  sscanf(argv[9],  "%lf", &nly1_s); 
  sscanf(argv[10], "%lf", &nlx2_s); 
  sscanf(argv[11], "%lf", &nly2_s); 
  sscanf(argv[12], "%lf", &alpha);
  sscanf(argv[13], "%lf", &beta_start);
  sscanf(argv[14], "%lf", &beta_end);
  sscanf(argv[15], "%lf", &sigma);
  sscanf(argv[16], "%lf", &delta_param);
  sscanf(argv[17], "%lf", &t_until_stationary);
  iterations = (unsigned long) round(t_end/dt);
  if (beta_end == beta_start) {
    no_paramvalues = 1;
  }
  else {
    no_paramvalues = (unsigned long) round( (beta_end - beta_start)/delta_param) + 1;
  }
  until_stationary = (unsigned long) round(t_until_stationary/dt);
  if (DEBUG) {
    fprintf(stderr,"\nArguments for <%s>\n\n",argv[0]);
    fprintf(stderr,  "user_filenamesize:  %8d\n",  user_filenamesize);
    fprintf(stderr,  "user_filename:      <%s>\n", user_filename);
    fprintf(stderr,  "seed:               %8ld\n", gauss_seed);
    fprintf(stderr,  "t_end:              %8g\n",  t_end);
    fprintf(stderr,  "dt:                 %8g\n",  dt);
    fprintf(stderr,  "lx_s:               %8g\n",  lx_s); 
    fprintf(stderr,  "ly_s:               %8g\n",  ly_s); 
    fprintf(stderr,  "nlx1_s:             %8g\n",  nlx1_s); 
    fprintf(stderr,  "nly1_s:             %8g\n",  nly1_s); 
    fprintf(stderr,  "nlx2_s:             %8g\n",  nlx2_s); 
    fprintf(stderr,  "nly2_s:             %8g\n",  nly2_s); 
    fprintf(stderr,  "alpha:              %8g\n",  alpha);
    fprintf(stderr,  "beta_start:         %8g\n",  beta_start);
    fprintf(stderr,  "beta_end:           %8g\n",  beta_end);
    fprintf(stderr,  "sigma:              %8g\n",  sigma);
    fprintf(stderr,  "delta_param:        %8g\n",  delta_param);
    fprintf(stderr,  "t_until_stationary: %8g\n",  t_until_stationary);
    fprintf(stderr,  "no_paramvalues:     %8lu\n",  no_paramvalues);
    fprintf(stderr,  "until_stationary:   %8lu\n\n",  until_stationary);
  }
}

void save_data_in_matlab_format (char          *filename_mat,
				 double        *l1vector,
				 double        *l2vector, 
				 double        *pvector,
				 unsigned long no_param
				 ) {
  FILE     *fp;  
  double   tmp;

  if ((fp = fopen(filename_mat,"wb")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_mat);fflush(stderr);
    exit(1); 
  }
  savemat(fp,TYPE_FLAG, "alpha",                   1,1,0,        &alpha,               (double *) NULL);    
  savemat(fp,TYPE_FLAG, "beta_low",                1,1,0,        &beta_start,          (double *) NULL);
  savemat(fp,TYPE_FLAG, "beta_high",               1,1,0,        &beta_end,            (double *) NULL);
  savemat(fp,TYPE_FLAG, "sigma",                   1,1,0,        &sigma,               (double *) NULL); 
  savemat(fp,TYPE_FLAG, "time_stepsize",           1,1,0,        &dt,                  (double *) NULL); 
  savemat(fp,TYPE_FLAG, "beta_values_V",           1,no_param,0, pvector,              (double *) NULL); 
  savemat(fp,TYPE_FLAG, "first_lyap_exp_V",        1,no_param,0, l1vector,             (double *) NULL); 
  savemat(fp,TYPE_FLAG, "second_lyap_exp_V",       1,no_param,0, l2vector,             (double *) NULL); 
  tmp = log10(t_end);
  savemat(fp,TYPE_FLAG, "log10_t_end",             1,1,0,        &tmp,                 (double *) NULL); 
  tmp = (double ) gauss_seed;
  savemat(fp,TYPE_FLAG, "seed",                    1,1,0,        &tmp,                 (double *) NULL); 
  savemat(fp,TYPE_FLAG, "t_until_stationary",      1,1,0,        &t_until_stationary,  (double *) NULL);
  savemat(fp,TYPE_FLAG, "parameter_stepsize",      1,1,0,        &delta_param,         (double *) NULL);
  tmp = (double ) no_param;
  savemat(fp,TYPE_FLAG, "no_parameter_values",     1,1,0,        &tmp,                 (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_linearized_x",         1,1,0,        &lx_s,                (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_linearized_y",         1,1,0,        &ly_s,                (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_nonlinear_x1",         1,1,0,        &nlx1_s,              (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_nonlinear_y1",         1,1,0,        &nly1_s,              (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_nonlinear_x2",         1,1,0,        &nlx2_s,              (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_nonlinear_y2",         1,1,0,        &nly2_s,              (double *) NULL);
  fclose(fp);
}

void save_data_in_ascii_format (char           *filename_asc,
				double         *l1vector,
				double         *l2vector, 
				double         *pvector,
				unsigned long  no_param
				) {
  FILE           *fp;
  unsigned long  i;

  if ((fp = fopen(filename_asc,"wt")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_asc);fflush(stderr);
    exit(1); 
  }
  fprintf(fp,  "## Parameters:\n##\n");
  fprintf(fp,  "#alpha=                                   %8g\n",  alpha);
  fprintf(fp,  "#beta_low=                                %8g\n",  beta_start);
  fprintf(fp,  "#beta_high=                               %8g\n",  beta_end);
  fprintf(fp,  "#sigma=                                   %8g\n",  sigma);
  fprintf(fp,  "#time_stepsize=                           %8g\n",  dt);
  fprintf(fp,  "#t_end=                                   %8g\n",  t_end);
  fprintf(fp,  "#parameter_stepsize=                      %8g\n",  delta_param);
  fprintf(fp,  "#no_parameter_values=                     %8lu\n", no_paramvalues);
  fprintf(fp,  "#initial_value_for_linearized_system_x=   %8g\n",  lx_s); 
  fprintf(fp,  "#initial_value_for_linearized_system_y=   %8g\n",  ly_s); 
  fprintf(fp,  "#initial_value_for_nonlinear_system_x1=   %8g\n",  nlx1_s); 
  fprintf(fp,  "#initial_value_for_nonlinear_system_y1=   %8g\n",  nly1_s); 
  fprintf(fp,  "#initial_value_for_nonlinear_system_x2=   %8g\n",  nlx2_s); 
  fprintf(fp,  "#initial_value_for_nonlinear_system_y2=   %8g\n",  nly2_s); 
  fprintf(fp,  "#seed=                                    %8ld\n", gauss_seed);
  fprintf(fp,  "#time_to_reach_stationary_solution=       %8g\n",  t_until_stationary);
  fprintf(fp,  "\n\n\n## Lyapunov exponents:\n##\n");
  fprintf(fp,  "##%6s\t%s\t%s\n\n", "beta",
	  "first_lyapunov_exp(beta)",
	  "second_lyapunov_exp(beta)");
  for (i=0;i<no_param;i++) {
    fprintf(fp,"%8g\t\t\t%8g\t\t\t%8g \n", pvector[i],l1vector[i],l2vector[i]); 
  }
  fclose(fp);
}

int main(int argc, char *argv[]) {

  unsigned long  i,j;
  char           filename_mat[NAMESIZE],filename_asc[NAMESIZE];
  char           filename_log[NAMESIZE];
  FILE           *fp_log;
  double         lx_old=0.0,nlx_old=0.0;
  double         lx_new=0.0,nlx_new=0.0;
  double         ly_old=0.0,nly_old=0.0;
  double         ly_new=0.0,nly_new=0.0;
  double         nlx_stationary,nly_stationary;
  double         *lyap_exp1_vector;
  double         *lyap_exp2_vector;
  double         *pvector,*sde_vector;
  double         time;
  double         lyap_exp1_tmp=0.0,lyap_exp2_tmp=0.0;
  double         trapsum=0.0;
  double         bruss_fwd_trace_Df_constant_part;
  int            modulus;
  int            overflow_flag=0;
  
  testenv(argv[0]);
  
  initialize(argv);
  if ( MODULUS_TIME>0 ) {
    modulus = (int) round(MODULUS_TIME/dt);
  } else {
    modulus = 1/dt;
  }
  sprintf(filename_mat, "%s.mat", user_filename);
  sprintf(filename_asc, "%s.dat", user_filename);
  sprintf(filename_log, "%s.log", user_filename);
  if ((fp_log = fopen(filename_log,"wt")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_log);
    exit(1); 
  }
  sde_vector = salloc(sizeof(*sde_vector)*SDE_DIMENSION);
  pvector = salloc(sizeof(*pvector)*no_paramvalues);
  lyap_exp1_vector = salloc(sizeof(*lyap_exp1_vector)*no_paramvalues);
  lyap_exp2_vector = salloc(sizeof(*lyap_exp2_vector)*no_paramvalues);
  for (i=0;i<no_paramvalues;i++) {
    lyap_exp1_vector[i] = MAX_SDE;
    lyap_exp2_vector[i] = MAX_SDE;
  }
  beta = beta_start;
  for(i=0; i<no_paramvalues; i++) {
    bruss_fwd_trace_Df_constant_part = -(beta + 1);
    /* compute stationary nonlinear solution */
    compute_stationary_nl_solution(alpha,beta,sigma,dt,
				   nlx1_s,nlx2_s,nly1_s,nly2_s,
				   fp_log,modulus,
				   &nlx_stationary,&nly_stationary);
/*
    if (DEBUG) {
      fprintf(stderr, "\n\nComputing the Lyapunov exponents for BETA=%g:\n\n", beta);
    }
*/
    gauss_init(gauss_seed);
    lx_old = lx_s;
    ly_old = ly_s;
    nlx_old = nlx_stationary;
    nly_old = nly_stationary;
    trapsum =  bruss_fwd_trace_Df_timedep_part(nlx_old,nly_old) * 0.5;
    time = dt;
    overflow_flag=0;
    for(j=0;j<iterations;j++) { 
      fill_gauss(&gauss1);
      compute_j_functionals(&gauss1,dt);	
      Brus_linear_ord_2(lx_old, nlx_old, &lx_new, 
			ly_old, nly_old, &ly_new, 
			dt, beta, sigma); 
      Brus_nonlinear_ord_2(nlx_old, &nlx_new, nly_old, &nly_new, 
			   dt, alpha, beta, sigma); 
      if ( (fabs(nlx_new) > MAX_SDE) || (fabs(nly_new) > MAX_SDE) ) {
	fprintf(fp_log, "Computation of the nonlinear solution terminated\n");
	fprintf(fp_log, "at time = %g for beta = %g and the initial value (x(0),y(0))=(%g,%g)\n", 
		time,beta,nlx_stationary,nly_stationary);
        fprintf(fp_log, "as |x_nonlinear(%g)| or |y_nonlinear(%g)| > %g\n",time,time,MAX_SDE);
	lyap_exp1_vector[i] = MAX_SDE;
	lyap_exp2_vector[i] = MAX_SDE;
	overflow_flag=1;
	break;
      }
      if ( (fabs(lx_new) > MAX_SDE) || (fabs(ly_new) > MAX_SDE) ) {
	fprintf(fp_log, "Computation of the linear solution terminated\n");
	fprintf(fp_log, "at time = %g for beta = %g and the initial value (x(0),y(0))=(%g,%g)\n", 
		time,beta,lx_s,ly_s);
	fprintf(fp_log, "as |x_linear(%g)| or |y_linear(%g)| > %g\n",time,time,MAX_SDE);
	lyap_exp1_vector[i] = MAX_SDE;
	lyap_exp2_vector[i] = MAX_SDE;
	overflow_flag=1;
	break;
      }
      trapsum +=  bruss_fwd_trace_Df_timedep_part(nlx_new,nly_new) * 0.5;
      sde_vector[0] = lx_new;
      sde_vector[1] = ly_new;
      lyap_exp1_tmp = log(norm_euklid(sde_vector,SDE_DIMENSION));
      lyap_exp1_tmp /= time;
      lyap_exp2_tmp = bruss_fwd_trace_Df_constant_part 
	              + ( trapsum / ((double) (j+1)) )
	              - lyap_exp1_tmp;
/*
      if ( DEBUG ) {
	if ( ((j+1) % modulus )==0 ) {
	  fprintf(stderr,"\ntime=%g\n",time);
	  fprintf(stderr,"lx:      %g \t ly:      %g\n",lx_new,ly_new);
	  fprintf(stderr,"lyap_exp1:      %g \t lyap_exp2:      %g\n\n",lyap_exp1_tmp,lyap_exp2_tmp);
	}
      }
*/
      nlx_old = nlx_new;
      nly_old = nly_new;
      lx_old  = lx_new;
      ly_old  = ly_new;
      time += dt;
      trapsum +=  bruss_fwd_trace_Df_timedep_part(nlx_new,nly_new) * 0.5;
    }
    if (!overflow_flag) {
      lyap_exp1_vector[i] = lyap_exp1_tmp;
      lyap_exp2_vector[i] = lyap_exp2_tmp;
    }
    pvector[i] = beta;
    beta += delta_param;
  }
  save_data_in_matlab_format (filename_mat,lyap_exp1_vector,lyap_exp2_vector, 
			      pvector,no_paramvalues);
  save_data_in_ascii_format (filename_asc,lyap_exp1_vector,lyap_exp2_vector, 
			     pvector,no_paramvalues);
  fclose(fp_log);

  return 0;
}







