#include <stdlib.h>
#include <stdio.h> 
#include <string.h>
#include <math.h>
#include <values.h>

#include "oned_slv.h"
#include "j.h"
#include "utils.h"

/*****  global variables  *****/

long            gauss_seed;
double          beta,beta_start,beta_end; 
double          sigma;
unsigned long   iterations;
double          t_end, t_until_stationary;
double          dt;                    
double          lx_s;
double          nlx1_s,nlx2_s;
double          **xnon;
unsigned long   no_paramvalues;
double          delta_param;
char            *user_filename;
int             user_filenamesize;

/*****  global variables  *****/

void initialize(char *argv[]) {

  sscanf(argv[1],  "%d",  &user_filenamesize);
  user_filename = salloc((user_filenamesize+1)*sizeof(*user_filename));
  sscanf(argv[2],  "%s",  user_filename);
  sscanf(argv[3],  "%ld", &gauss_seed);
  sscanf(argv[4],  "%lf", &t_end);
  sscanf(argv[5],  "%lf", &dt);
  sscanf(argv[6],  "%lf", &lx_s); 
  sscanf(argv[7],  "%lf", &nlx1_s); 
  sscanf(argv[8],  "%lf", &nlx2_s); 
  sscanf(argv[9],  "%lf", &beta_start);
  sscanf(argv[10], "%lf", &beta_end);
  sscanf(argv[11], "%lf", &sigma);
  sscanf(argv[12], "%lf", &delta_param);
  sscanf(argv[13], "%lf", &t_until_stationary);
  iterations = (unsigned long) round(t_end/dt);
  if (beta_end == beta_start) {
    no_paramvalues = 1;
  } else {
      no_paramvalues = (unsigned) round( (beta_end - beta_start)/delta_param) + 1;
  }
  beta = beta_start;
  until_stationary = (unsigned long) round(t_until_stationary/dt);
  if (DEBUG) {
    fprintf(stderr,  "\nArguments for <%s>\n\n",     argv[0]);
    fprintf(stderr,  "user_filenamesize:  %8d\n",    user_filenamesize);
    fprintf(stderr,  "user_filename:      <%s>\n",   user_filename);
    fprintf(stderr,  "seed=               %8ld\n",   gauss_seed);
    fprintf(stderr,  "t_end=              %8g\n",    t_end);
    fprintf(stderr,  "dt=                 %8g\n",    dt);
    fprintf(stderr,  "lx_s=               %8g\n",    lx_s); 
    fprintf(stderr,  "nlx1_s=             %8g\n",    nlx1_s); 
    fprintf(stderr,  "nlx2_s=             %8g\n",    nlx2_s); 
    fprintf(stderr,  "beta_start=         %8g\n",    beta_start);
    fprintf(stderr,  "beta_end=           %8g\n",    beta_end);
    fprintf(stderr,  "sigma=              %8g\n",    sigma);
    fprintf(stderr,  "delta_param=        %8g\n",    delta_param);
    fprintf(stderr,  "t_until_stationary= %8g\n",    t_until_stationary);
    fprintf(stderr,  "no_paramvalues:     %8lu\n",   no_paramvalues);
    fprintf(stderr,  "until_stationary:   %8lu\n\n", until_stationary);
  }
}

void save_data_in_matlab_format (char          *filename_mat,
				 double        *l1vector,
				 double        *pvector,
				 unsigned long no_param
				 ) {
  FILE     *fp;  
  double   tmp;

  if ((fp = fopen(filename_mat,"wb")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_mat);fflush(stderr);
    exit(1); 
  }
  savemat(fp,TYPE_FLAG, "beta_low",             1,1,0,         &beta_start,          (double *) NULL);
  savemat(fp,TYPE_FLAG, "beta_high",            1,1,0,         &beta_end,            (double *) NULL);
  savemat(fp,TYPE_FLAG, "sigma",                1,1,0,         &sigma,               (double *) NULL); 
  savemat(fp,TYPE_FLAG, "time_stepsize",        1,1,0,         &dt,                  (double *) NULL); 
  savemat(fp,TYPE_FLAG, "beta_values_V",        1,no_param,0,  pvector,              (double *) NULL); 
  savemat(fp,TYPE_FLAG, "lyapunov_exponent_V",  1,no_param,0,  l1vector,             (double *) NULL); 
  tmp = log10(t_end);
  savemat(fp,TYPE_FLAG, "log10_t_end",          1,1,0,         &tmp,                 (double *) NULL); 
  tmp = (double ) gauss_seed;
  savemat(fp,TYPE_FLAG, "seed",                 1,1,0,         &tmp,                 (double *) NULL); 
  savemat(fp,TYPE_FLAG, "t_until_stationary",   1,1,0,         &t_until_stationary,  (double *) NULL); 
  savemat(fp,TYPE_FLAG, "parameter_stepsize",   1,1,0,         &delta_param,         (double *) NULL);
  tmp = (double ) no_param;
  savemat(fp,TYPE_FLAG, "no_parameter_values",  1,1,0,         &tmp,                 (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_linearized_x",      1,1,0,         &lx_s,                (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_nonlinear_x1",      1,1,0,         &nlx1_s,              (double *) NULL);
  savemat(fp,TYPE_FLAG, "iv_nonlinear_x2",      1,1,0,         &nlx2_s,              (double *) NULL);
  fclose(fp);
}

void save_data_in_ascii_format (char           *filename_asc,
				double         *l1vector,
				double         *pvector,
				unsigned long  no_param
				) {
  FILE           *fp;
  unsigned long  i;

  if ((fp = fopen(filename_asc,"wt")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_asc);fflush(stderr);
    exit(1); 
  }
  fprintf(fp,  "## Parameters:\n\n");
  fprintf(fp,  "#beta_low=                                %8g\n",  beta_start);
  fprintf(fp,  "#beta_high=                               %8g\n",  beta_end);
  fprintf(fp,  "#sigma=                                   %8g\n",  sigma);
  fprintf(fp,  "#time_stepsize=                           %8g\n",  dt);
  fprintf(fp,  "#t_end=                                   %8g\n",  t_end);
  fprintf(fp,  "#parameter_stepsize=                      %8g\n",  delta_param);
  fprintf(fp,  "#no_parameter_values=                     %8lu\n", no_paramvalues);
  fprintf(fp,  "#initial_value_for_linearized_system_x=   %8g\n",  lx_s); 
  fprintf(fp,  "#initial_value_for_nonlinear_system_x1=   %8g\n",  nlx1_s); 
  fprintf(fp,  "#initial_value_for_nonlinear_system_x2=   %8g\n",  nlx2_s); 
  fprintf(fp,  "#seed=                                    %8ld\n", gauss_seed);
  fprintf(fp,  "#time_to_reach_stationary_solution=       %8g\n",  t_until_stationary);
  fprintf(fp,  "\n\n\n## Lyapunov exponent:\n\n");
  fprintf(fp,  "##%6s\t%s\n\n", "beta", " lyapunov_exponent(beta)");
  for (i=0;i<no_param;i++) {
    fprintf(fp,"%8g\t\t\t%8g\n", pvector[i], l1vector[i]); 
  }
  fclose(fp);
}

int main(int argc, char *argv[]) {

  unsigned long   i,j;
  char            filename_mat[NAMESIZE],filename_asc[NAMESIZE];
  char            filename_log[NAMESIZE];
  FILE            *fp_log;
  double          lx_old=0.0,nlx_old=0.0;
  double          lx_new=0.0,nlx_new=0.0;
  double          nlx_stationary;
  double          *lyap_exp1_vector;
  double          *pvector;
  double          time;
  double          lyap_exp1_tmp=0.0;
  int             modulus;
  int             overflow_flag=0;
  
  testenv(argv[0]);
  
  initialize(argv);
  if ( MODULUS_TIME>0 ) {
    modulus = (int) round(MODULUS_TIME/dt);
  } else {
    modulus = 1/dt;
  }
  sprintf(filename_mat, "%s.mat", user_filename);
  sprintf(filename_asc, "%s.dat", user_filename);
  sprintf(filename_log, "%s.log", user_filename);
  if ((fp_log = fopen(filename_log,"wt")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_log);fflush(stderr);
    exit(1); 
  }
  pvector          = salloc(sizeof(*pvector)         *no_paramvalues);
  lyap_exp1_vector = salloc(sizeof(*lyap_exp1_vector)*no_paramvalues);
  for (i=0;i<no_paramvalues;i++) {
    lyap_exp1_vector[i] = MAX_SDE;
  }
  beta = beta_start;
  for(i=0;i<no_paramvalues;i++) {
    /* compute stationary nonlinear solution */
    compute_stat_nl_solution_oned (beta,sigma,dt,
				   nlx1_s,nlx2_s,
				   fp_log,modulus,
				   &nlx_stationary);
    gauss_init(gauss_seed);
    lx_old = lx_s;
    nlx_old = nlx_stationary;
    time = dt;
    overflow_flag=0;
    for(j=0; j<iterations;j++) {  
      fill_gauss(&gauss1);
      compute_j_functionals_oned(&gauss1,dt);	
      oned_linear_ord2_fwd (lx_old, nlx_old, &lx_new, dt, beta); 
      oned_nonlinear_ord2_fwd (nlx_old, &nlx_new, dt, beta, sigma); 
      if ( fabs(nlx_new) > MAX_SDE ) {
	fprintf(fp_log, "Computation of the nonlinear solution terminated\n");
	fprintf(fp_log, "at time = %g for beta = %g and the initial value x(0)=%g\n", 
		time,beta,nlx_stationary);
        fprintf(fp_log, "as |x_nonlinear(%g)| > %g\n",time,MAX_SDE);
	lyap_exp1_vector[i] = MAX_SDE;
	overflow_flag=1;
	break;
      }
      if ( fabs(lx_new) > MAX_SDE ) {
	fprintf(fp_log, "Computation of the linear solution terminated\n");
	fprintf(fp_log, "at time = %g for beta = %g and the initial value x(0)=%g\n", 
		time,beta,lx_s);
	fprintf(fp_log, "as |x_linear(%g)| > %g\n",time,MAX_SDE);
	lyap_exp1_vector[i] = MAX_SDE;
	overflow_flag=1;
	break;
      }
      lyap_exp1_tmp = log(fabs(lx_new));
      lyap_exp1_tmp /= time;
/*
      if ( DEBUG ) {
	if ( ((j+1) % modulus )==0 ) {
	  fprintf(stderr,"\ntime=%g\n",time);
	  fprintf(stderr,"lx:      %g\n",lx_new);
	  fprintf(stderr,"Lyap1:   %+.8f\n",lyap_exp1_tmp);
	  fflush(stderr);
	}
      }
*/
      nlx_old = nlx_new;
      lx_old  = lx_new;
      time += dt;
    }
    if (!overflow_flag) {
      lyap_exp1_vector[i] = lyap_exp1_tmp;
    }
    pvector[i] = beta;
    beta += delta_param;
  }
  save_data_in_matlab_format(filename_mat,lyap_exp1_vector, 
			     pvector,no_paramvalues);
  save_data_in_ascii_format(filename_asc,lyap_exp1_vector, 
			    pvector,no_paramvalues);
  fclose(fp_log);

  return 0;
}








