#include <stdlib.h>
#include <stdio.h> 
#include <string.h>
#include <math.h>
#include <values.h>

#include "utils.h"
#include "j.h"
#include "brus_slv.h"


/*****  global variables  *****/

long            gauss_seed;
double          alpha;
double          beta,beta_start,beta_end;
double          sigma;
double          time;
double          t_end; 
unsigned long   iterations;
double          t_until_stationary;
double          dt;                    
double          lx_s, ly_s;
double          nlx1_s,nly1_s,nlx2_s,nly2_s;
unsigned long   no_paramvalues;
double          delta_param;
char            *user_filename;
int             user_filenamesize;

/*****  global variables  *****/

void initialize(char *argv[]) {

  sscanf(argv[1],   "%d",   &user_filenamesize);
  user_filename = salloc((user_filenamesize+1)*sizeof(*user_filename));
  sscanf(argv[2],   "%s",   user_filename);
  sscanf(argv[3],   "%ld",  &gauss_seed);
  sscanf(argv[4],   "%lf",  &t_end);
  sscanf(argv[5],   "%lf",  &dt);
  sscanf(argv[6],   "%lf",  &lx_s); 
  sscanf(argv[7],   "%lf",  &ly_s); 
  sscanf(argv[8],   "%lf",  &nlx1_s); 
  sscanf(argv[9],   "%lf",  &nly1_s); 
  sscanf(argv[10],  "%lf",  &nlx2_s); 
  sscanf(argv[11],  "%lf",  &nly2_s); 
  sscanf(argv[12],  "%lf",  &alpha);
  sscanf(argv[13],  "%lf",  &beta_start);
  sscanf(argv[14],  "%lf",  &beta_end);
  sscanf(argv[15],  "%lf",  &sigma);
  sscanf(argv[16],  "%lf",  &delta_param);
  sscanf(argv[17],  "%lf",  &t_until_stationary);
  iterations = (unsigned long) round(t_end/dt);
  if (beta_end == beta_start) {
    no_paramvalues = 1;
  } else {
    no_paramvalues = (unsigned long) round( (beta_end - beta_start)/delta_param) + 1;
  }
  until_stationary = (unsigned long) round(t_until_stationary/dt);
  if (DEBUG) {
    fprintf(stderr,"\nArguments for <%s>\n\n",argv[0]);
    fprintf(stderr,  "user_filenamesize: %8d\n",    user_filenamesize);
    fprintf(stderr,  "user_filename:     <%s>\n",   user_filename);
    fprintf(stderr,  "seed:              %8ld\n",   gauss_seed);
    fprintf(stderr,  "t_end:             %8g\n",    t_end);
    fprintf(stderr,  "dt:                %8g\n",    dt);
    fprintf(stderr,  "lx_s:              %8g\n",    lx_s); 
    fprintf(stderr,  "ly_s:              %8g\n",    ly_s); 
    fprintf(stderr,  "nlx1_s:            %8g\n",    nlx1_s); 
    fprintf(stderr,  "nly1_s:            %8g\n",    nly1_s); 
    fprintf(stderr,  "nlx2_s:            %8g\n",    nlx2_s); 
    fprintf(stderr,  "nly2_s:            %8g\n",    nly2_s); 
    fprintf(stderr,  "alpha:             %8g\n",    alpha);
    fprintf(stderr,  "beta_start:        %8g\n",    beta_start);
    fprintf(stderr,  "beta_end:          %8g\n",    beta_end);
    fprintf(stderr,  "sigma:             %8g\n",    sigma);
    fprintf(stderr,  "delta_param:       %8g\n",    delta_param);
    fprintf(stderr,  "t_until_stationary %8g\n",    t_until_stationary);
    fprintf(stderr,  "no_paramvalues:    %8lu\n",   no_paramvalues);
    fprintf(stderr,  "until_stationary:  %8lu\n\n", until_stationary);
  }
}

void save_data_in_matlab_format (char            *filename_mat, 
				 double          *pvector,
				 double          *no_rp_vector,
				 double          *no_rn_vector,
				 double          *angle_vector,
				 unsigned long   no_param
				 ) {
  double  tmp;
  FILE    *fp_mat;
  
  fp_mat = fopen(filename_mat,"wb");
  if (fp_mat == NULL) {
    fprintf(stderr,"Unable to open file %s\n",filename_mat);fflush(stderr);
    exit(1);
  }
  savemat(fp_mat,TYPE_FLAG, "alpha",               1,1,0,        &alpha,              (double *) NULL);    
  savemat(fp_mat,TYPE_FLAG, "beta_start",          1,1,0,        &beta_start,         (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "beta_end",            1,1,0,        &beta_end,           (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "sigma",               1,1,0,        &sigma,              (double *) NULL); 
  savemat(fp_mat,TYPE_FLAG, "time_stepsize",       1,1,0,        &dt,                 (double *) NULL); 
  savemat(fp_mat,TYPE_FLAG, "beta_values_V",       1,no_param,0, pvector,             (double *) NULL); 
  savemat(fp_mat,TYPE_FLAG, "no_pos_rotations_V",  1,no_param,0, no_rp_vector,        (double *) NULL); 
  savemat(fp_mat,TYPE_FLAG, "no_neg_rotations_V",  1,no_param,0, no_rn_vector,        (double *) NULL); 
  savemat(fp_mat,TYPE_FLAG, "rotation_numbers_V",  1,no_param,0, angle_vector,        (double *) NULL); 
  tmp = log10(t_end);
  savemat(fp_mat,TYPE_FLAG, "log10_t_end",         1,1,0,        &tmp,                (double *) NULL); 
  tmp = (double ) gauss_seed;
  savemat(fp_mat,TYPE_FLAG, "seed",                1,1,0,        &tmp,                (double *) NULL); 
  tmp = (double ) no_param;
  savemat(fp_mat,TYPE_FLAG, "no_parameter_values", 1,1,0,        &tmp,                (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "t_until_stationary",  1,1,0,        &t_until_stationary, (double *) NULL); 
  savemat(fp_mat,TYPE_FLAG, "parameter_stepsize",  1,1,0,        &delta_param,        (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "iv_linearized_x",     1,1,0,        &lx_s,               (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "iv_linearized_y",     1,1,0,        &ly_s,               (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "iv_nonlinear_x1",     1,1,0,        &nlx1_s,             (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "iv_nonlinear_y1",     1,1,0,        &nly1_s,             (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "iv_nonlinear_x2",     1,1,0,        &nlx2_s,             (double *) NULL);
  savemat(fp_mat,TYPE_FLAG, "iv_nonlinear_y2",     1,1,0,        &nly2_s,             (double *) NULL);
  fclose(fp_mat);
}

void save_data_in_ascii_format (char            *filename_asc, 
				double          *pvector,
				double          *no_rp_vector,
				double          *no_rn_vector,
				double          *angle_vector,
				unsigned long   no_param
				) {
  FILE            *fp_asc;
  unsigned long   i;
  
  fp_asc = fopen(filename_asc,"wt");
  if (fp_asc == NULL) {
    fprintf(stderr,"Unable to open file %s\n",filename_asc);fflush(stderr);
    exit(1);
  }
  fprintf(fp_asc,  "## Parameters: \n##\n");
  fprintf(fp_asc,  "#alpha=                                   %8g\n",  alpha);
  fprintf(fp_asc,  "#beta_low=                                %8g\n",  beta_start);
  fprintf(fp_asc,  "#beta_high=                               %8g\n",  beta_end);
  fprintf(fp_asc,  "#sigma=                                   %8g\n",  sigma);
  fprintf(fp_asc,  "#time_stepsize=                           %8g\n",  dt);
  fprintf(fp_asc,  "#t_end=                                   %8g\n",  t_end);
  fprintf(fp_asc,  "#parameter_stepsize=                      %8g\n",  delta_param);
  fprintf(fp_asc,  "#no_parameter_values=                     %8lu\n", no_paramvalues);
  fprintf(fp_asc,  "#initial_value_for_linearized_system_x=   %8g\n",  lx_s); 
  fprintf(fp_asc,  "#initial_value_for_linearized_system_y=   %8g\n",  ly_s); 
  fprintf(fp_asc,  "#initial_value_for_nonlinear_system_x1=   %8g\n",  nlx1_s); 
  fprintf(fp_asc,  "#initial_value_for_nonlinear_system_y1=   %8g\n",  nly1_s); 
  fprintf(fp_asc,  "#initial_value_for_nonlinear_system_x2=   %8g\n",  nlx2_s); 
  fprintf(fp_asc,  "#initial_value_for_nonlinear_system_y2=   %8g\n",  nly2_s); 
  fprintf(fp_asc,  "#seed=                                    %8ld\n", gauss_seed);
  fprintf(fp_asc,  "#time_to_reach_stationary_solution=       %8g\n\n",  t_until_stationary);
  fprintf(fp_asc,"## Rotation number:\n##\n"); 
  fprintf(fp_asc,"## number of positive rotations, number of negative rotations, rotation number:\n\n"); 
  fprintf(fp_asc,"##%6s\t%s\t%s\t%s\n\n","beta","no_pos_rot(beta)",
	  "no_neg_rot(beta)","   rotation_number(beta)" ); 
  for (i=0;i<no_param;i++) {
    fprintf(fp_asc,"%8g\t\t%8g\t\t%8g\t\t\t%8g\n",pvector[i],
	    no_rp_vector[i],no_rn_vector[i],angle_vector[i]); 
  }
  fclose(fp_asc);
}

int main(int argc, char *argv[]) {

  unsigned long   i,j;
  char            filename_mat[NAMESIZE],filename_asc[NAMESIZE];
  char            filename_log[NAMESIZE];
  FILE            *fp_log;
  double          lx_old=0.0,nlx_old=0.0;
  double          lx_new=0.0,nlx_new=0.0;
  double          ly_old=0.0,nly_old=0.0; 
  double          ly_new=0.0,nly_new=0.0;
  double          nlx_stationary,nly_stationary;
  double          *pvector,*no_rotpos_vector,*no_rotneg_vector;
  double          *angle_sum_vector;
  int             modulus;
  int             ly_changed;
  int             rot_pos,rot_neg;
  double          angle_sum,vector_product_z;
  double          cos_phi,start_x,start_y;
  int             overflow_flag=0;
  
  testenv(argv[0]);

  initialize(argv);
  if ( MODULUS_TIME>0 ) {
    modulus = (int) round(MODULUS_TIME/dt);
  } else {
    modulus = 1/dt;
  }
  sprintf(filename_mat, "%s.mat", user_filename);
  sprintf(filename_asc, "%s.dat", user_filename);
  sprintf(filename_log, "%s.log", user_filename);
  fp_log = fopen(filename_log,"wt");
  if (fp_log == NULL) {
    fprintf(stderr,"Unable to open file <%s>.\n",filename_log);fflush(stderr);
    exit(1);
  }
  pvector          = salloc(sizeof(*pvector)         *no_paramvalues);
  no_rotpos_vector = salloc(sizeof(*no_rotpos_vector)*no_paramvalues);
  no_rotneg_vector = salloc(sizeof(*no_rotneg_vector)*no_paramvalues);
  angle_sum_vector = salloc(sizeof(*angle_sum_vector)*no_paramvalues);
  for (i=0;i<no_paramvalues;i++) {
    no_rotpos_vector[i]  = 0;
    no_rotneg_vector[i]  = 0;
    angle_sum_vector[i]  = 0;
  }
  beta = beta_start;
  for(i=0; i<no_paramvalues; i++) {
    rot_pos = rot_neg = 0;
    ly_changed=0;
    angle_sum=0.0;
    cos_phi = 1;
    /* compute stationary nonlinear solution */
    compute_stationary_nl_solution (alpha,beta,sigma,dt,
				    nlx1_s,nlx2_s,nly1_s,nly2_s,
				    fp_log,modulus,
				    &nlx_stationary,&nly_stationary);
    gauss_init(gauss_seed);
    start_x = lx_old = lx_s;
    start_y = ly_old = ly_s;
    nlx_old = nlx_stationary;
    nly_old = nly_stationary;
/*
    if (DEBUG) {
	fprintf(stderr,"%g\t%g\t%g\t# time, lx, ly\n",0.0,lx_old,ly_old);
    }
*/
    time = dt;
    overflow_flag=0;
    for(j=0;j<iterations;j++) { 
      fill_gauss(&gauss1);
      compute_j_functionals(&gauss1,dt);	
      Brus_linear_ord_2(lx_old, nlx_old, &lx_new, 
			ly_old, nly_old, &ly_new, 
			dt, beta, sigma); 
      Brus_nonlinear_ord_2(nlx_old, &nlx_new, nly_old, &nly_new, 
			   dt, alpha, beta, sigma); 
      if ( (fabs(nlx_new) > MAX_SDE) || (fabs(nly_new) > MAX_SDE) ) {
	fprintf(fp_log, "Computation of the nonlinear solution terminated\n");
	fprintf(fp_log, "at time = %g for beta = %g and the initial value (x(0),y(0))=(%g,%g)\n", 
		time,beta,nlx_stationary,nly_stationary);
        fprintf(fp_log, "as |x_nonlinear(%g)| or |y_nonlinear(%g)| > %g\n",time,time,MAX_SDE);
	no_rotpos_vector[i] = MAX_SDE;
	no_rotneg_vector[i] = MAX_SDE;
	angle_sum_vector[i] = MAX_SDE;
	overflow_flag=1;
	break;
      }
      if ( (fabs(lx_new) > MAX_SDE) || (fabs(ly_new) > MAX_SDE) ) {
	fprintf(fp_log, "Computation of the linear solution terminated\n");
	fprintf(fp_log, "at time = %g for beta = %g and the initial value (x(0),y(0))=(%g,%g)\n", 
		time,beta,lx_s,ly_s);
	fprintf(fp_log, "as |x_linear(%g)| or |y_linear(%g)| > %g\n",time,time,MAX_SDE);
	no_rotpos_vector[i] = MAX_SDE;
	no_rotneg_vector[i] = MAX_SDE;
	angle_sum_vector[i] = MAX_SDE;
	overflow_flag=1;
	break;
      }
      nlx_old = nlx_new;
      nly_old = nly_new;
/*
      if (DEBUG) {
	if ( ((j+1) % modulus )==0 ) {
	  fprintf(stderr,"%g\t%g\t%g\t# time, lx, ly\n",time,lx_new,ly_new);
	}
      }
*/
      cos_phi = (start_x*lx_new + start_y*ly_new)
	        /(sqrt((start_x*start_x + start_y*start_y)
	               *(lx_new*lx_new + ly_new*ly_new)));
      if (fabs(cos_phi) < 0.5) {
	vector_product_z = start_x*ly_new - start_y*lx_new;
	if (vector_product_z >0) {
	  angle_sum += acos(cos_phi);
	} else {
	  angle_sum -= acos(cos_phi);
	}
	start_x = lx_new;
	start_y = ly_new;
/*
	if (DEBUG) {
	  fprintf(stderr,"angle_segment: %8g \t vector_product_z: %8g\n",
		  acos(cos_phi),vector_product_z);fflush(stderr);
	}
*/
	cos_phi = 1;
      }
      if ((ly_old*ly_new) <= 0) {
	ly_changed++;
      }
      if ( (lx_old*lx_new <= 0) && ( (ly_changed&1) == 1) ) {
	ly_changed = 0;
	if (lx_old < lx_new) {
	  if (ly_old > 0) {
	    rot_neg++;
	  } else {
	    rot_pos++;
	  }
	} else {
	  if (ly_old > 0) {
	    rot_pos++;
	  } else {
	    rot_neg++;
	  }
	}
      }
      lx_old = lx_new;
      ly_old = ly_new;
      time += dt;
    }
    pvector[i] = beta;
    beta += delta_param;
    if (!overflow_flag) {
      no_rotpos_vector[i] = rot_pos*M_PI/t_end;
      no_rotneg_vector[i] = rot_neg*M_PI/t_end;
      vector_product_z = start_x*ly_new - start_y*lx_new;
      if (vector_product_z >0) {
	angle_sum_vector[i] = (angle_sum+acos(cos_phi))/t_end;
      } else {
	angle_sum_vector[i] = (angle_sum-acos(cos_phi))/t_end;
      }
    }
/*
    if (DEBUG) {
      fprintf(stderr,"angle_segment: %8g \t vector_product_z: %8g\n",
	      acos(cos_phi),vector_product_z);fflush(stderr);
      fprintf(stderr, "no_rotpos[%ld]: %8g \t no_rotneg[%ld]: %8g \t angle_sum[%ld]: %g\n", 
	              i,no_rotpos_vector[i],i,no_rotneg_vector[i],i,angle_sum_vector[i]);
    }
*/
  }
  save_data_in_matlab_format (filename_mat,pvector,no_rotpos_vector,no_rotneg_vector,
			 angle_sum_vector,no_paramvalues);
  save_data_in_ascii_format (filename_asc,pvector,no_rotpos_vector,no_rotneg_vector,
			 angle_sum_vector,no_paramvalues);
  fclose(fp_log);

  return 0;
}








