#include <stdlib.h>
#include <stdio.h> 
#include <string.h>
#include <math.h>
#include <values.h>

#include "utils.h"
#include "j.h"
#include "brus_slv.h"

/******  global variables  ******/

long             gauss_seed;
double           alpha, beta, sigma;
unsigned long    iterations;
double           dt,t_end;                    
double           x_s,y_s;
double           *x_trj, *y_trj;
double           *times_vector;
char             *user_filename;
int              user_filenamesize;

/******  global variables  ******/

void initialize(char *argv[]) {

  sscanf(argv[1],  "%d",   &user_filenamesize);
  user_filename = salloc((user_filenamesize+1)*sizeof(*user_filename));
  sscanf(argv[2],  "%s",   user_filename);
  sscanf(argv[3],  "%ld",  &gauss_seed);
  sscanf(argv[4],  "%lf",  &t_end); 
  sscanf(argv[5],  "%lf",  &dt);
  sscanf(argv[6],  "%lf",  &x_s); 
  sscanf(argv[7],  "%lf",  &y_s); 
  sscanf(argv[8],  "%lf",  &alpha);
  sscanf(argv[9],  "%lf",  &beta);
  sscanf(argv[10], "%lf",  &sigma);
  iterations = (unsigned long) round(t_end / dt);
  if (DEBUG) {
    fprintf(stderr,"\nArguments for <%s>\n\n",argv[0]);
    fprintf(stderr,  "user_filenamesize:   %8d\n",    user_filenamesize);
    fprintf(stderr,  "user_filename:      <%s>\n",    user_filename);
    fprintf(stderr,  "gauss_seed:          %8ld\n",   gauss_seed);
    fprintf(stderr,  "t_end:               %8g\n",    t_end);
    fprintf(stderr,  "dt:                  %8g\n",    dt);
    fprintf(stderr,  "x_s:                 %8g\n",    x_s);
    fprintf(stderr,  "y_s:                 %8g\n",    y_s);
    fprintf(stderr,  "alpha:               %8g\n",    alpha);
    fprintf(stderr,  "beta:                %8g\n",    beta);
    fprintf(stderr,  "sigma:               %8g\n",    sigma);
    fprintf(stderr,  "iterations:          %8lu\n",   iterations);
  }
  x_trj   =  (double*) salloc( (iterations+1)*sizeof(*x_trj));
  y_trj   =  (double*) salloc( (iterations+1)*sizeof(*y_trj));
  times_vector = (double*) salloc( (iterations+1)*sizeof(*times_vector));
}

void save_data_in_matlab_format(char  *filename_mat) {

  FILE           *fp;
  double         tmp;
  unsigned long  i;

  if ((fp = fopen(filename_mat,"wb")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n",filename_mat);fflush(stderr);
    exit(1);  
  }
  for (i=0;i<(iterations+1);i++) {
    times_vector[i] *=  -1;
  }
  tmp= log10((double) t_end);
  savemat(fp,TYPE_FLAG, "log10_t_end",     1,1,0,             &tmp,         (double *) NULL);
  savemat(fp,TYPE_FLAG, "x_trajectory_V",  1,iterations+1,0,  x_trj,        (double *) NULL); 
  savemat(fp,TYPE_FLAG, "y_trajectory_V",  1,iterations+1,0,  y_trj,        (double *) NULL); 
  savemat(fp,TYPE_FLAG, "times_V",         1,iterations+1,0,  times_vector, (double *) NULL); 
  savemat(fp,TYPE_FLAG, "alpha",           1,1,0,             &alpha,       (double *) NULL);
  savemat(fp,TYPE_FLAG, "beta",            1,1,0,             &beta,        (double *) NULL);
  savemat(fp,TYPE_FLAG, "sigma",           1,1,0,             &sigma,       (double *) NULL); 
  savemat(fp,TYPE_FLAG, "time_stepsize",   1,1,0,             &dt,          (double *) NULL);
  savemat(fp,TYPE_FLAG, "initial_value_x", 1,1,0,             &x_s,         (double *) NULL);
  savemat(fp,TYPE_FLAG, "initial_value_y", 1,1,0,             &y_s,         (double *) NULL);
  tmp = (double ) gauss_seed;
  savemat(fp,TYPE_FLAG, "seed",            1,1,0,             &tmp,         (double *) NULL); 
  fclose(fp);
}

int main(int argc, char *argv[]) {

  unsigned long     k;
  FILE              *fp_asc,*fp_log;
  char              filename_mat[NAMESIZE];
  char              filename_asc[NAMESIZE],filename_log[NAMESIZE];
  double            x,y;
  
  testenv(argv[0]);
  
  initialize(argv);
  sprintf(filename_mat, "%s.mat", user_filename);
  sprintf(filename_asc, "%s.dat", user_filename);
  sprintf(filename_log, "%s.log", user_filename);
  if ((fp_asc = fopen(filename_asc,"wt")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_asc);fflush(stderr);
    exit(1);  
  }
  if ((fp_log = fopen(filename_log,"wt")) == NULL) {
    fprintf(stderr,"Unable to open file %s\n", filename_log);fflush(stderr);
    exit(1);  
  }

  fprintf(fp_asc,"## Parameters:\n\n");
  fprintf(fp_asc,"#alpha=           %8g\n",  alpha);
  fprintf(fp_asc,"#beta=            %8g\n",  beta);
  fprintf(fp_asc,"#sigma=           %8g\n",  sigma);
  fprintf(fp_asc,"#t_end=           %8g\n",  t_end);
  fprintf(fp_asc,"#time_stepsize=   %8g\n",  dt);
  fprintf(fp_asc,"#initial_value_x= %8g\n",  x_s);
  fprintf(fp_asc,"#initial_value_y= %8g\n",  y_s);
  fprintf(fp_asc,"#seed=            %8ld\n", gauss_seed);
  fprintf(fp_asc,"\n\n");

  gauss_init(gauss_seed);
  times_vector[0] = 0.0;
  x_trj[0] = x = x_s;
  y_trj[0] = y = y_s;

  fprintf(fp_asc,"## Trajectory (x(t),y(t)):\n##\n");
  fprintf(fp_asc,"##%6s\t%8s\t%8s\n", "t", "x(t)", "y(t)");
  fprintf(fp_asc,"%8g\t%8g\t%8g\n", times_vector[0], x_trj[0], y_trj[0]);

  for(k=0;k<iterations;k++) {
    fill_gauss(&gauss1);
    compute_j_functionals(&gauss1,dt);
    times_vector[k+1] = times_vector[k] + dt;
    if ( (x != MAX_SDE) && (y != MAX_SDE) ) {
      Brus_nonlinear_ord2_bwd_for_grid (&x,&y,dt,alpha,beta,sigma);
      x_trj[k+1] = x; 
      y_trj[k+1] = y; 
    }
    if ( (fabs(x_trj[k+1]) > MAX_SDE) || (fabs(y_trj[k+1]) > MAX_SDE) ) {
      fprintf(fp_log, "Computation of the trajectory terminated\n"); 
      fprintf(fp_log, "at time = %g for beta = %g and the initial value (x(0),y(0))=(%g,%g)\n", 
	      -times_vector[k+1],beta,x_s,y_s);
      fprintf(fp_log, "as |x(%g)| or |y(%g)| > %g\n",-times_vector[k+1],-times_vector[k+1],MAX_SDE);
      x_trj[k+1] = MAX_SDE;
      y_trj[k+1] = MAX_SDE;
      break;
    }
    fprintf(fp_asc,"%8g\t%8g\t%8g\n", -times_vector[k+1], x_trj[k+1], y_trj[k+1]);
  }
  save_data_in_matlab_format(filename_mat);
  fclose(fp_asc);
  fclose(fp_log);

  return 0;
}









