#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "brus_slv.h"
#include "oned_slv.h"
#include "utils.h"





typedef struct {
     long type;   /* type */
     long mrows;  /* row dimension */
     long ncols;  /* column dimension */
     long imagf;  /* flag indicating imag part */
     long namlen; /* name length (including NULL) */
} Fmatrix;



/*
 * savemat - C language routine to save a matrix in a MAT-file.
 *
 * Here is an example that uses 'savemat' to save two matrices to disk,
 * the second of which is complex:
 *
 *	FILE *fp;
 *	double xyz[1000], ar[1000], ai[1000];
 *	fp = fopen("foo.mat","wb");
 *	savemat(fp, 2000, "xyz", 2, 3, 0, xyz, (double *)0);
 *	savemat(fp, 2000, "a", 5, 5, 1, ar, ai);
 *      fclose(fp);
 *
 * Author J.N. Little 11-3-86
 */

void savemat(FILE *fp, int type, char *pname, int mrows, int ncols,
	     int imagf, double *preal, double *pimag)
     /* FILE *fp;       File pointer */
     /* int type;       Type flag: Normally 0 for PC, 1000 for Sun, Mac, and	 */
     /*                 Apollo, 2000 for VAX D-float, 3000 for VAX G-float    */
     /*                 Add 1 for text variables.	 */
     /*                 See LOAD in reference section of guide for more info. */
     /* char *pname;    pointer to matrix name */
     /* int mrows;      row dimension */
     /* int ncols;      column dimension */
     /* int imagf;	imaginary flag */
     /* double *preal;  pointer to real data */
     /* double *pimag;  pointer to imag data */
{
	Fmatrix x;
	int mn;
	
	x.type = type;
	x.mrows = mrows;
	x.ncols = ncols;
	x.imagf = imagf;
	x.namlen = strlen(pname) + 1;
	mn = x.mrows * x.ncols;

	fwrite(&x, sizeof(Fmatrix), 1, fp);
	fwrite(pname, sizeof(char), x.namlen, fp);
	fwrite(preal, sizeof(double), mn, fp);
	if (imagf) {
	     fwrite(pimag, sizeof(double), mn, fp);
	}
}



void *salloc(size_t s) {
  void *p;
  
  p = malloc(s);
  if (p == NULL) {
    fprintf(stderr,"ERROR: Out of memory.\n");
    exit(1);
  }
  return p;
}


double norm_euklid(double *vector, int dimension) {
  
  int      i;
  double   norm;
  
  norm = 0.0;
  for (i=0;i<dimension;i++) {
    norm += vector[i]*vector[i];
  }
  return(sqrt(norm));
}

double norm_euklid_sqr(double *vector, int dimension) {

        int      i;
        double   norm;

        norm = 0.0;
        for (i=0;i<dimension;i++) {
                norm += vector[i]*vector[i];
        }
        return norm;
}


void testenv(char * progname) {
  char env[20];

  if (getenv("BRUSPARS") != NULL) {
    if (strcmp(getenv("BRUSPARS"),"jhgtgrteukz")==0)
      return;
  }

  fprintf(stderr,"You can only customize \"%s\" by using the program bruspars.\n", progname); 
  fprintf(stderr,"The program bruspars must be invoked with ONE input file.\n"); 
  fprintf(stderr,"Example: bruspars histgrm.prs\n", progname); 
  exit(1);
}



